#include "board.h"
#include "mw.h"

extern float VelNorth, VelEast;  //Crashpilot
static void GPS_NewData(uint16_t c);
static void GPS_set_pids(void);
static void gpsPrint(const char *str);

const uint32_t init_speed[5] = { 9600, 19200, 38400, 57600, 115200 };

static const uint8_t ubloxInit[] = {
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x05, 0x00, 0xFF, 0x19,                                  // disable all default NMEA messages
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x03, 0x00, 0xFD, 0x15,
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x01, 0x00, 0xFB, 0x11,
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x00, 0x00, 0xFA, 0x0F,
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x02, 0x00, 0xFC, 0x13,
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0xF0, 0x04, 0x00, 0xFE, 0x17,
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0x01, 0x02, 0x01, 0x0E, 0x47,                                  // set POSLLH MSG rate
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0x01, 0x03, 0x01, 0x0F, 0x49,                                  // set STATUS MSG rate
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0x01, 0x06, 0x01, 0x12, 0x4F,                                  // set SOL MSG rate
     0xB5, 0x62, 0x06, 0x01, 0x03, 0x00, 0x01, 0x12, 0x01, 0x1E, 0x67,                                  // set VELNED MSG rate
     0xB5, 0x62, 0x06, 0x16, 0x08, 0x00, 0x03, 0x07, 0x03, 0x00, 0x51, 0x08, 0x00, 0x00, 0x8A, 0x41,    // set WAAS to EGNOS
     0xB5, 0x62, 0x06, 0x08, 0x06, 0x00, 0xC8, 0x00, 0x01, 0x00, 0x01, 0x00, 0xDE, 0x6A                 // set rate to 5Hz
};

  #define MTK_BAUD_RATE_57600			"$PMTK251,57600*2C\r\n"
  #define MTK_SBAS_INTEGRITYMODE	"$PMTK319,1*24\r\n"
  #define MTK_OUTPUT_5HZ					"$PMTK220,200*2C\r\n"
  #define MTK_NAVTHRES_OFF      	"$PMTK397,0*23\r\n"
  #define MTK_SBAS_ON							"$PMTK313,1*2E\r\n"
  #define MTK_WAAS_ON           	"$PMTK301,2*2E\r\n"
  #define MTK_SET_BINARY					"$PGCMD,16,0,0,0,0,0*6A\r\n"

void gpsInit(uint32_t baudrate)                               // Called in Main
{
  uint8_t i;
	uint32_t timeout;

  GPS_set_pids();
  timeout = micros()+10000000;					                      // 10 sec timeout
	GPS_Present = 0;
	delay(1000);                                                // let it init
	while (GPS_Present == 0){                                   // Repeat while no GPS Data
		if (micros()>timeout) break;                              // Stop that after timeout
    uart2Init(baudrate, GPS_NewData, false);
  	switch (cfg.gps_type){ 	                                  // GPS_NMEA = 0, GPS_UBLOX = 1, GPS_MTK16 = 2, GPS_MTK19 = 3,
      case 0:                                                 // GPS_NMEA
      break;
      case 1:                                                 // GPS_UBLOX
        for (i = 0; i < 5; i++) {
          uart2ChangeBaud(init_speed[i]);
          switch (baudrate){
            case 19200:
              gpsPrint("$PUBX,41,1,0003,0001,19200,0*23\r\n");
            break;
            case 38400:
              gpsPrint("$PUBX,41,1,0003,0001,38400,0*26\r\n");
            break;
            case 57600:
              gpsPrint("$PUBX,41,1,0003,0001,57600,0*2D\r\n");
            break;
            case 115200:
              gpsPrint("$PUBX,41,1,0003,0001,115200,0*1E\r\n");
            break;
          }
        delay(10);
        }
        uart2ChangeBaud(baudrate);
        for (i = 0; i < sizeof(ubloxInit); i++){
          uart2Write(ubloxInit[i]);                          // send ubx init binary
          delay(4);
        }
      break;
      case 2:                                                // GPS_MTK16
      case 3:                                                // GPS_MTK19
        for (i = 0; i < 5; i++){
          uart2ChangeBaud(init_speed[i]);
          delay(200);
          gpsPrint(MTK_BAUD_RATE_57600);    
        }
 		    uart2ChangeBaud(57600);
        delay(200);
	      gpsPrint(MTK_SET_BINARY);
        delay(200);
        gpsPrint(MTK_OUTPUT_5HZ);
        delay(200);
        gpsPrint(MTK_SBAS_INTEGRITYMODE);
        delay(200);
        gpsPrint(MTK_NAVTHRES_OFF);
        delay(200);
        gpsPrint(MTK_SBAS_ON);
        delay(200);
        gpsPrint(MTK_WAAS_ON);
      break;
    }
    delay(1000);
  }
  if (GPS_Present) sensorsSet(SENSOR_GPS);                // Do we get Data?
}

static void gpsPrint(const char *str)
{
    while (*str) {
        uart2Write(*str);
        if (cfg.gps_type == 1)                            // GPS_NMEA = 0, GPS_UBLOX = 1, GPS_MTK16 = 2, GPS_MTK19 = 3,
            delay(5);
        str++;
    }
    // wait to send all
    while (!uart2TransmitEmpty());
    delay(30);
}

/*-----------------------------------------------------------
 *
 * Multiwii GPS code - revision: 1187
 *
 *-----------------------------------------------------------*/
#define POSHOLD_IMAX           20       // degrees
#define POSHOLD_RATE_IMAX      20       // degrees
#define NAV_IMAX               20       // degrees

/* GPS navigation can control the heading */
#define NAV_TAIL_FIRST             0    // true - copter comes in with tail first
#define NAV_SET_TAKEOFF_HEADING    1    // true - when copter arrives to home position it rotates it's head to takeoff direction

#define GPS_FILTERING              1    // add a 5 element moving average filter to GPS coordinates, helps eliminate gps noise but adds latency
//#define GPS_LOW_SPEED_D_FILTER     1    // below .5m/s speed ignore D term for POSHOLD_RATE, theoretically this also removed D term induced noise

static bool check_missed_wp(void);
static void GPS_distance_cm_bearing(int32_t * lat1, int32_t * lon1, int32_t * lat2, int32_t * lon2, uint32_t * dist, int32_t * bearing);
//static void GPS_distance(int32_t lat1, int32_t lon1, int32_t lat2, int32_t lon2, uint16_t* dist, int16_t* bearing);
static void GPS_calc_longitude_scaling(int32_t lat);
static void GPS_calc_velocity(void);
static void GPS_calc_location_error(int32_t * target_lat, int32_t * target_lng, int32_t * gps_lat, int32_t * gps_lng);
static void GPS_calc_poshold(void);
static void GPS_calc_nav_rate(int16_t max_speed);
static void GPS_update_crosstrack(void);
static bool GPS_newFrame(char c);
static bool GPS_NMEA_newFrame(char c);
static bool GPS_MTK_newFrame(uint8_t data); 
static bool GPS_UBLOX_newFrame(uint8_t data);
static bool UBLOX_parse_gps(void);
static int16_t GPS_calc_desired_speed(int16_t max_speed, bool _slow);
int32_t wrap_18000(int32_t error);
static int32_t wrap_36000(int32_t angle);

typedef struct PID_PARAM_ { 							// mwii1320
  float kP;
  float kI;
  float kD;
  float Imax;
} PID_PARAM;

static PID_PARAM posholdPID_PARAM;  			// mwii1320
static PID_PARAM poshold_ratePID_PARAM;
static PID_PARAM navPID_PARAM;

typedef struct PID_ {  									 // mwii1320
    float   integrator; // integrator value
    int32_t last_input; // last input for derivative
    float   lastderivative; // last derivative for low-pass filter
    float   output;
    float   derivative;
} PID;

static PID posholdPID[2];  							// mwii1320
static PID poshold_ratePID[2];
static PID navPID[2];

static int32_t get_P(int32_t error, struct PID_PARAM_* pid)  //mwii1320
{
    return (float)error * pid->kP;
}

static int32_t get_I(int32_t error, float* dt, struct PID_* pid, struct PID_PARAM_* pid_param)
{
    pid->integrator += ((float)error * pid_param->kI) * *dt;
    pid->integrator = constrain(pid->integrator,-pid_param->Imax,pid_param->Imax);
    return pid->integrator;
}

static int32_t get_D(int32_t input, float* dt, struct PID_* pid, struct PID_PARAM_* pid_param) { // dt in milliseconds //mwii1320
    pid->derivative = (input - pid->last_input) / *dt;
    float filter = 7.9577e-3;
    pid->derivative = pid->lastderivative + (*dt / ( filter + *dt)) * (pid->derivative - pid->lastderivative);
    pid->last_input = input;
    pid->lastderivative    = pid->derivative;
    return pid_param->kD * pid->derivative;
  }

//static int32_t get_D(int32_t input, float *dt, PID *pid, PID_PARAM *pid_param)
//#define PID_FILTER       (1.0f / (2.0f * M_PI * (float)cfg.gps_lpf))

static void reset_PID(struct PID_* pid) //mwii1320
{
    pid->integrator = 0;
    pid->last_input = 0;
    pid->lastderivative = 0;
}

#define GPS_X 1
#define GPS_Y 0

#define CROSSTRACK_GAIN            1      // #define NAV_SPEED_MIN 100 // cm/sec // #define NAV_SPEED_MAX 300 // cm/sec
#define NAV_SLOW_NAV               true
#define NAV_BANK_MAX 3000                 // 30deg max banking when navigating (just for security and testing)
static float   dTnav;                     // Delta Time in milliseconds for navigation computations, updated with every good GPS read
static int16_t actual_speed[2] = {0,0};
static float   CosLatScaleLon;          // this is used to offset the shrinking longitude as we go towards the poles
static int16_t rate_error[2];             // The difference between the desired rate of travel and the actual rate of travel
static int32_t error[2];                  // updated after GPS read - 5-10hz
static int32_t GPS_WP[2];                 // Currently used WP

// Location & Navigation
static int32_t target_bearing;            // This is the angle from the copter to the "next_WP" location in degrees * 100
// Crosstrack
static int32_t original_target_bearing;   // deg * 100, The original angle to the next_WP when the next_WP was set Also used to check when we pass a WP
static int16_t crosstrack_error;          // The amount of angle correction applied to target_bearing to bring the copter back on its optimum path
////////////////////////////////////////////////////////////////////////////////
// The location of the copter in relation to home, updated every GPS read (1deg - 100)
// static int32_t home_to_copter_bearing; /* unused */
// distance between plane and home in cm
// static int32_t home_distance; /* unused */
// distance between plane and next_WP in cm
static uint32_t wp_distance;
static uint16_t waypoint_speed_gov;       // used for slow speed wind up when start navigation;
static int32_t  nav_bearing;              // This is the angle from the copter to the "next_WP" location  with the addition of Crosstrack error in degrees * 100
static int16_t  nav_takeoff_bearing;      // saves the bearing at takeof (1deg = 1) used to rotate to takeoff direction when arrives at home

void GPS_NewData(uint16_t c)              // Called by uart2Init interrupt
{
    static   uint32_t nav_loopTimer;      // uint8_t axis; //Compilerwarning unused var
    uint32_t dist;
    int32_t  dir;
    int16_t  speed;
    static   int32_t LatSpikeTab[5];      // Crashpilot Spikefilter
    static   int32_t LonSpikeTab[5];      // Crashpilot Spikefilter
  	int32_t  extmp;                       // Crashpilot Spikefilter
    uint8_t  rdy;                         // Crashpilot Spikefilter
		uint8_t  sortidx;                     // Crashpilot Spikefilter
		uint8_t  maxsortidx;                  // Crashpilot Spikefilter
	
    if (GPS_newFrame(c)) {
        if (GPS_update == 1) GPS_update = 0;
        else GPS_update = 1;
			
        if (f.GPS_FIX && GPS_numSat >= 5){// Do gps stuff with at least 5 Sats
            if (!f.ARMED) f.GPS_FIX_HOME = 0;
            if (!f.GPS_FIX_HOME && f.ARMED) GPS_reset_home_position();
            extmp = GPS_coord[LAT];       //  Crashpilot Spikefilter START
            LatSpikeTab[4] = extmp;
            LatSpikeTab[0] = extmp;
            extmp = GPS_coord[LON];
            LonSpikeTab[4] = extmp;
            LonSpikeTab[0] = extmp;
            rdy = 0; maxsortidx=4;
            while(rdy == 0){
             rdy = 1;
             for (sortidx = 0; sortidx<maxsortidx;sortidx++){
              extmp = LatSpikeTab[sortidx];
              if (extmp > LatSpikeTab[sortidx+1]) {LatSpikeTab[sortidx] = LatSpikeTab[sortidx+1]; LatSpikeTab[sortidx+1] = extmp; rdy = 0;}
             }
            maxsortidx --;}
            rdy = 0; maxsortidx=4;
            while(rdy == 0){
             rdy = 1;
             for (sortidx = 0; sortidx<maxsortidx;sortidx++){
              extmp = LonSpikeTab[sortidx];
              if (extmp > LonSpikeTab[sortidx+1]) {LonSpikeTab[sortidx] = LonSpikeTab[sortidx+1]; LonSpikeTab[sortidx+1] = extmp; rdy = 0;}
             }
            maxsortidx --;}
            GPS_coord[LAT] = LatSpikeTab[2];
            GPS_coord[LON] = LonSpikeTab[2];//  Crashpilot Spikefilter END

						// dTnav calculation Time for calculating x,y speed and navigation pids
            dTnav = (float) (millis() - nav_loopTimer) / 1000.0f;
            nav_loopTimer = millis();
            // prevent runup from bad GPS
            dTnav = min(dTnav, 1.0f);

            // calculate distance and bearings for gui and other stuff continously - From home to copter
            GPS_distance_cm_bearing(&GPS_coord[LAT], &GPS_coord[LON], &GPS_home[LAT], &GPS_home[LON], &dist, &dir);
            GPS_distanceToHome = dist / 100;
            GPS_directionToHome = dir / 100;
		
            if (!f.GPS_FIX_HOME) {      // If we don't have home set, do not display anything
                GPS_distanceToHome = 0;
                GPS_directionToHome = 0;
            }

            // calculate the current velocity based on gps coordinates continously to get a valid speed at the moment when we start navigating
            GPS_calc_velocity();

            if (f.GPS_HOLD_MODE || f.GPS_HOME_MODE) { // ok we are navigating
                // do gps nav calculations here, these are common for nav and poshold
                GPS_distance_cm_bearing(&GPS_coord[LAT], &GPS_coord[LON], &GPS_WP[LAT], &GPS_WP[LON], &wp_distance, &target_bearing);
                GPS_calc_location_error(&GPS_WP[LAT], &GPS_WP[LON], &GPS_coord[LAT], &GPS_coord[LON]);

                switch (nav_mode) {
                case NAV_MODE_POSHOLD:
                    // Desired output is in nav_lat and nav_lon where 1deg inclination is 100
                    GPS_calc_poshold();
                    break;

                case NAV_MODE_WP:
                    speed = GPS_calc_desired_speed(cfg.nav_speed_max, NAV_SLOW_NAV);    // slow navigation
                    // use error as the desired rate towards the target
                    // Desired output is in nav_lat and nav_lon where 1deg inclination is 100
                    GPS_calc_nav_rate(speed);

                    // Tail control
                    if (cfg.nav_controls_heading){
                        if (NAV_TAIL_FIRST) magHold = wrap_18000(nav_bearing - 18000) / 100;
                         else magHold = nav_bearing / 100;
                    }
                    // Are we there yet ?(within x meters of the destination)
                    if ((wp_distance <= cfg.gps_wp_radius) || check_missed_wp()){      // if yes switch to poshold mode
                        nav_mode = NAV_MODE_POSHOLD;
                        if (NAV_SET_TAKEOFF_HEADING) magHold = nav_takeoff_bearing;
                    }
                    break;
                }               // END of gps calcs i.e navigating
            }                   // END of gps calcs i.e navigating
        }                       // END GPS_numSat >= 5
    }                           // END GPS_newFrame
}                               // END of void

void GPS_reset_home_position(void)
{
    if (f.GPS_FIX && GPS_numSat >= 5) {
        GPS_home[LAT] = GPS_coord[LAT];
        GPS_home[LON] = GPS_coord[LON];
        GPS_calc_longitude_scaling(GPS_coord[LAT]); // need an initial value for distance and bearing calc
        nav_takeoff_bearing = heading;              // save takeoff heading
        // Set ground altitude
        f.GPS_FIX_HOME = 1;
    }
}

//reset navigation (stop the navigation processor, and clear nav)
void GPS_reset_nav(void)
{
    uint8_t i;
    for (i = 0; i < 2; i++) {
        GPS_angle[i] = 0;
        nav_rated[i] = 0;
        nav[i] = 0;
        reset_PID(&posholdPID[i]);
        reset_PID(&poshold_ratePID[i]);
        reset_PID(&navPID[i]);
    }
}

//Get the relevant P I D values and set the PID controllers
static void GPS_set_pids(void)
{
    posholdPID_PARAM.kP = (float)cfg.P8[PIDPOS] / 100.0f;
    posholdPID_PARAM.kI = (float)cfg.I8[PIDPOS] / 100.0f;
    posholdPID_PARAM.Imax = POSHOLD_RATE_IMAX * 100;

    poshold_ratePID_PARAM.kP = (float)cfg.P8[PIDPOSR] / 10.0f;
    poshold_ratePID_PARAM.kI = (float)cfg.I8[PIDPOSR] / 100.0f;
    poshold_ratePID_PARAM.kD = (float)cfg.D8[PIDPOSR] / 1000.0f;
    poshold_ratePID_PARAM.Imax = POSHOLD_RATE_IMAX * 100;

    navPID_PARAM.kP = (float)cfg.P8[PIDNAVR] / 10.0f;
    navPID_PARAM.kI = (float)cfg.I8[PIDNAVR] / 100.0f;
    navPID_PARAM.kD = (float)cfg.D8[PIDNAVR] / 1000.0f;
    navPID_PARAM.Imax = POSHOLD_RATE_IMAX * 100;
}

// OK here is the onboard GPS code

////////////////////////////////////////////////////////////////////////////////////
// PID based GPS navigation functions
// Author : EOSBandi
// Based on code and ideas from the Arducopter team: Jason Short,Randy Mackay, Pat Hickey, Jose Julio, Jani Hirvinen
// Andrew Tridgell, Justin Beech, Adam Rivera, Jean-Louis Naudin, Roberto Navoni


////////////////////////////////////////////////////////////////////////////////////
// Sets the waypoint to navigate, reset neccessary variables and calculate initial values
//
void GPS_set_next_wp(int32_t *lat, int32_t *lon)
{
    GPS_WP[LAT] = *lat;
    GPS_WP[LON] = *lon;

    GPS_calc_longitude_scaling(*lat);
    GPS_distance_cm_bearing(&GPS_coord[LAT], &GPS_coord[LON], &GPS_WP[LAT], &GPS_WP[LON], &wp_distance, &target_bearing);

    nav_bearing = target_bearing;
    GPS_calc_location_error(&GPS_WP[LAT], &GPS_WP[LON], &GPS_coord[LAT], &GPS_coord[LON]);
    original_target_bearing = target_bearing;
    waypoint_speed_gov = cfg.nav_speed_min;
}
////////////////////////////////////////////////////////////////////////////////////
// Check if we missed the destination somehow
//
static bool check_missed_wp(void)
{
    int32_t temp;
    temp = target_bearing - original_target_bearing;
    temp = wrap_18000(temp);
    return (abs(temp) > 10000); // we passed the waypoint by 100 degrees
}

static void GPS_calc_longitude_scaling(int32_t lat)
{
	  double temp = lat;
    double rads = (abs(temp) / 10000000.0f) * RADX;
//    float rads = (abs((float)lat) / 10000000.0f) * 0.0174532925f;
    CosLatScaleLon = cosf(rads);
}

////////////////////////////////////////////////////////////////////////////////////
// Get distance between two points in cm
// Get bearing from pos1 to pos2, returns an 1deg = 100 precision
// http://www.movable-type.co.uk/scripts/latlong.html
// Distance formula:
// var x = (lon2-lon1) * Math.cos((lat1+lat2)/2);
// var y = (lat2-lat1);
// var d = Math.sqrt(x*x + y*y) * R;
// R = average earth radius 637100078,5 cm
// 
static void GPS_distance_cm_bearing(int32_t * lat1, int32_t * lon1, int32_t * lat2, int32_t * lon2, uint32_t * dist, int32_t * bearing)
{
	  double dradLatsq, dradLonsq, dradLat, dradLon, dLat, dLon;
	  if (*lat2 != 0 && *lat1 != 0 && *lon2 != 0 && *lon1 != 0){                     // Crashpilot Errorcheck
	      dLat = *lat2 - *lat1;                                                      // difference of latitude in 1/10 000 000 degrees
        dLon = *lon2 - *lon1;
			  dradLat = dLat / 10000000.0f * RADX;                                       // dlat in rad
			  dradLon = dLon / 10000000.0f * RADX * CosLatScaleLon;                      // dlon in rad and scale
			  dradLatsq = dradLat * dradLat;                                             // square values
			  dradLonsq = dradLon * dradLon;
        *dist = sqrtf(dradLatsq + dradLonsq) * 637100078.5f;                       // Earthradius cm
			  if (CosLatScaleLon !=0) dLat = dLat * 1/CosLatScaleLon;                    // Seen since apm 2.8.1
				*bearing = 9000.0f + atan2f(-dLat, dLon) * 5729.57795f;                    // Convert the output radians to 100xdeg
        if (*bearing < 0) *bearing += 36000;
		} else {                                                                       // Error!
			  *dist = 0;
			  *bearing = 0;
		}
}

////////////////////////////////////////////////////////////////////////////////////
// Calculate our current speed vector from gps position data
//
static void GPS_calc_velocity(void)
{
    static int16_t speed_old[2] = { 0, 0 };
    static int32_t last[2] = { 0, 0 };
    static uint8_t init = 0;
    static uint32_t lasttime = 0;
    uint32_t actualtime;
    // actual_speed[GPS_Y] y_GPS_speed positve = Up (NORTH)
    // actual_speed[GPS_X] x_GPS_speed positve = Right (EAST)
    float tmp = 1.0f / dTnav;
    if (init){
        actual_speed[GPS_X] = (float)(GPS_coord[LON] - last[LON]) * 1.11194940f * tmp * CosLatScaleLon; // To rad, include Earthradius and Time
 			  actual_speed[GPS_Y] = (float)(GPS_coord[LAT] - last[LAT]) * 1.11194940f * tmp;                  // To rad, include Earthradius and Time
			  actual_speed[GPS_X] = (actual_speed[GPS_X] + speed_old[GPS_X]) / 2;
        actual_speed[GPS_Y] = (actual_speed[GPS_Y] + speed_old[GPS_Y]) / 2;
        speed_old[GPS_X] = actual_speed[GPS_X];
        speed_old[GPS_Y] = actual_speed[GPS_Y];
			  actualtime = micros();
			  if ((actualtime - lasttime) < 500000 && cfg.gps_ins_cf !=0) {   // Only accept summation if last correction is not older than 0,5s
			      VelNorth = VelNorth * cfg.gps_ins_cf + (float)actual_speed[GPS_Y] * (1.0f - cfg.gps_ins_cf);  // CF: GPS Correction
				    VelEast  = VelEast  * cfg.gps_ins_cf + (float)actual_speed[GPS_X] * (1.0f - cfg.gps_ins_cf);  // CF: GPS Correction
			      actual_speed[GPS_Y] = VelNorth;
			      actual_speed[GPS_X] = VelEast;
				} else {                                                  // INS Correction not possible: Timeout or turned off
				    VelNorth = 0;
					  VelEast = 0;
				}
				lasttime = actualtime;
    } else {
	      actual_speed[GPS_Y] = 0;
        actual_speed[GPS_X] = 0;
		}
    init = 1;
    last[LON] = GPS_coord[LON];
    last[LAT] = GPS_coord[LAT];
}
////////////////////////////////////////////////////////////////////////////////////
// Calculate a location error between two gps coordinates
// Because we are using lat and lon to do our distance errors here's a quick chart:
//      100     = 1m
//      1000    = 11m    = 36 feet
//      1800    = 19.80m = 60 feet
//      3000    = 33m
//      10000   = 111m
//
static void GPS_calc_location_error(int32_t *target_lat, int32_t *target_lng, int32_t *gps_lat, int32_t *gps_lng)
{
    error[LON] = (float) (*target_lng - *gps_lng) * 1.11194940f * CosLatScaleLon;   // X Error
    error[LAT] = (float) (*target_lat - *gps_lat) * 1.11194940f;                    // Y Error
}

////////////////////////////////////////////////////////////////////////////////////
// Calculate nav_lat and nav_lon from the x and y error and the speed
//
static void GPS_calc_poshold(void)
{
    int32_t d;
    int32_t target_speed;
    uint8_t axis;

    for (axis = 0; axis < 2; axis++) {
        target_speed = get_P(error[axis], &posholdPID_PARAM);       // calculate desired speed from lon error
        target_speed = constrain(target_speed,-100,100);
			  rate_error[axis] = target_speed - actual_speed[axis];       // calc the speed error

        nav[axis] = get_P(rate_error[axis], &poshold_ratePID_PARAM) +
                    get_I(rate_error[axis] + error[axis], &dTnav, &poshold_ratePID[axis], &poshold_ratePID_PARAM);
        d = get_D(error[axis], &dTnav, &poshold_ratePID[axis], &poshold_ratePID_PARAM);
        d = constrain(d, -2000, 2000);
        if (abs(actual_speed[axis]) < 50) d = 0;                  // get rid of noise
        nav[axis] +=d;
        nav[axis] = constrain(nav[axis], -NAV_BANK_MAX, NAV_BANK_MAX);
        navPID[axis].integrator = poshold_ratePID[axis].integrator;
    }
}
////////////////////////////////////////////////////////////////////////////////////
// Calculate the desired nav_lat and nav_lon for distance flying such as RTH
//
static void GPS_calc_nav_rate(int16_t max_speed)
{
    float trig[2];
    float temp;
    uint8_t axis;

    // push us towards the original track
    GPS_update_crosstrack();

    // nav_bearing includes crosstrack
    temp = (9000l - nav_bearing) * RADX100;
	  trig[GPS_X] = cosf(temp);
    trig[GPS_Y] = sinf(temp);

	
    for (axis = 0; axis < 2; axis++) {
        rate_error[axis] = (trig[axis] * max_speed) - actual_speed[axis];
        rate_error[axis] = constrain(rate_error[axis], -1000, 1000);
        // P + I + D
        nav[axis] = get_P(rate_error[axis], &navPID_PARAM) +
                    get_I(rate_error[axis], &dTnav, &navPID[axis], &navPID_PARAM) +
                    get_D(rate_error[axis], &dTnav, &navPID[axis], &navPID_PARAM);

        nav[axis] = constrain(nav[axis], -NAV_BANK_MAX, NAV_BANK_MAX);
        poshold_ratePID[axis].integrator = navPID[axis].integrator;
    }
}
////////////////////////////////////////////////////////////////////////////////////
// Calculating cross track error, this tries to keep the copter on a direct line
// when flying to a waypoint.
//
static void GPS_update_crosstrack(void)
{
    if (abs(wrap_18000(target_bearing - original_target_bearing)) < 4500) {     // If we are too far off or too close we don't do track following
        float temp = (target_bearing - original_target_bearing) * RADX100;
        crosstrack_error = sinf(temp) * (wp_distance * CROSSTRACK_GAIN); // Meters we are off track line
        nav_bearing = target_bearing + constrain(crosstrack_error, -3000, 3000);
        nav_bearing = wrap_36000(nav_bearing);
    } else {
        nav_bearing = target_bearing;
    }
}
////////////////////////////////////////////////////////////////////////////////////
// Determine desired speed when navigating towards a waypoint, also implement slow
// speed rampup when starting a navigation
//
//      |< WP Radius
//      0  1   2   3   4   5   6   7   8m
//      ...|...|...|...|...|...|...|...|
//                100  |  200     300     400cm/s
//                 |                                        +|+
//                 |< we should slow to 1.5 m/s as we hit the target
//
static int16_t GPS_calc_desired_speed(int16_t max_speed, bool _slow)
{
    // max_speed is default 400 or 4m/s
    if (_slow) {
        max_speed = min(max_speed, wp_distance / 2);
    } else {
        max_speed = min(max_speed, wp_distance);
        max_speed = max(max_speed, cfg.nav_speed_min);      // go at least 100cm/s
    }

    // limit the ramp up of the speed
    // waypoint_speed_gov is reset to 0 at each new WP command
    if (max_speed > waypoint_speed_gov) {
        waypoint_speed_gov += (int) (100.0f * dTnav);    // increase at .5/ms
        max_speed = waypoint_speed_gov;
    }
    return max_speed;
}
////////////////////////////////////////////////////////////////////////////////////
// Utilities
//
int32_t wrap_18000(int32_t error)
{
    if (error > 18000)
        error -= 36000;
    if (error < -18000)
        error += 36000;
    return error;
}

static int32_t wrap_36000(int32_t angle)
{
    if (angle > 36000)
        angle -= 36000;
    if (angle < 0)
        angle += 36000;
    return angle;
}

// This code is used for parsing NMEA data
#define DIGIT_TO_VAL(_x)    (_x - '0')
uint32_t GPS_coord_to_degrees(char* s)
{
    char *p, *q;
    uint8_t deg = 0, min = 0;
    unsigned int frac_min = 0;
    int i;

    // scan for decimal point or end of field
    for (p = s; isdigit(*p); p++);
    q = s;

    // convert degrees
    while ((p - q) > 2) {
        if (deg)
            deg *= 10;
        deg += DIGIT_TO_VAL(*q++);
    }
    // convert minutes
    while (p > q) {
        if (min)
            min *= 10;
        min += DIGIT_TO_VAL(*q++);
    }
    // convert fractional minutes
    // expect up to four digits, result is in
    // ten-thousandths of a minute
    if (*p == '.') {
        q = p + 1;
        for (i = 0; i < 4; i++) {
            frac_min *= 10;
            if (isdigit(*q))
                frac_min += *q++ - '0';
        }
    }
    return deg * 10000000UL + (min * 1000000UL + frac_min * 100UL) / 6;
}

// helper functions
static uint32_t grab_fields(char *src, uint8_t mult)
{                               // convert string to uint32
    uint8_t i;
    uint32_t tmp = 0;
    for (i = 0; src[i] != 0; i++){
        if (src[i] == '.'){
            i++;
            if (mult == 0) break;
            else src[i + mult] = 0;
        }
        tmp *= 10;
        if (src[i] >= '0' && src[i] <= '9')
            tmp += src[i] - '0';
    }
    return tmp;
}

static uint8_t hex_c(uint8_t n)
{                               // convert '0'..'9','A'..'F' to 0..15
    n -= '0';
    if (n > 9)
        n -= 7;
    n &= 0x0F;
    return n;
}

static bool GPS_newFrame(char c)                      // Crashpilot
{
  	switch (cfg.gps_type){                            // GPS_NMEA = 0, GPS_UBLOX = 1, GPS_MTK16 = 2, GPS_MTK19 = 3,
      case 0:                                         // NMEA
      return GPS_NMEA_newFrame(c);					
      case 1:                                         // UBX
      return GPS_UBLOX_newFrame(c);
			case 2:                                         // Dealing with old, faulty and new, correct binary protocol
			case 3: 
      return GPS_MTK_newFrame(c);                     // GPS_MTK_newFrame handles both 1.6 and 1.9 3drobotics nomenclature
    }
    return false;
}

static bool GPS_MTK_newFrame(uint8_t data){           // Crashpilot: This code is stupid but works
   static uint8_t pstep;                              // Parse Step
   static uint8_t lastbyte;                           // Last Byte for Sync detection
   static uint8_t LSLshifter;                         // Bitshiftvalue
   static uint8_t chkA,count;
   static int32_t lat;                                // MTK Dataset
   static int32_t lon;                                // MTK Dataset
   static int32_t alt;                                // MTK Dataset
   static int32_t grspeed;                            // MTK Dataset
   static int32_t grcourse;                           // MTK Dataset
   static uint8_t satn,fixtype;                       // MTK Dataset
   int32_t tmp32;
   uint8_t startbyte;                                 // must be unsigned
   bool parsed;
   parsed = false;
   startbyte = 0;                                     // Initialize to be on the safe side so casemachine will bail out on VERY unlikely error
   // GPS_NMEA = 0, GPS_UBLOX = 1, GPS_MTK16 = 2, GPS_MTK19 = 3,
	 if (cfg.gps_type == 2) startbyte = 0xd0;           // 3drobotics 1.6 FW and clones have $d0 preamblebyte no 1
	 if (cfg.gps_type == 3) startbyte = 0xd1;           // 3drobotics 1.9 FW and clones have $d1 preamblebyte no 1
 	 if (pstep == 0 && data == 0xdd && lastbyte == startbyte) pstep = 100; // Detect Sync "0xD1,0xDD" Only search for Sync when not already decoding
   lastbyte = data;
   switch(pstep) {
    case 0:                                           // Special Case: Do Nothing
    break;
    case 100:                                         // Special Case: Prepare next decoding run
    pstep = 1;                                        // Jump into decoding on next run
    chkA = 0; count = 0;                              // Reset values
    break;
    case 1:                                           // Payload Byte is always $20! (This is the first Byte after sync preamble)
    if (data == 0x20) pstep++;                        // Since it is always $20 we take it as extended, undocumented syncword "preamble3"
     else pstep =0;                                   // Error! Wait for sync
    chkA = chkA + data;
    count ++;
    break;
    case 2:                                           // Read Dataset Latitude
    lat = data;
    LSLshifter = 0;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 3:
    LSLshifter = LSLshifter+8;
    tmp32 = data;
    tmp32 = tmp32<<LSLshifter;
    lat = lat | tmp32;
    if (LSLshifter == 24) pstep++;                    // FW APM TEST //    if (LSLshifter == 24){lat = lat * 10; pstep++;}
    chkA = chkA + data;
    count ++;
    break;    
    case 4:                                           // Read Dataset Longitude
    lon = data;
    LSLshifter = 0;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 5:
    LSLshifter = LSLshifter+8;
    tmp32 = data;
    tmp32 = tmp32<<LSLshifter;
    lon = lon | tmp32;
    if (LSLshifter == 24) pstep++;                    // FW APM TEST //    if (LSLshifter == 24){lon = lon * 10; pstep++;}
    chkA = chkA + data;
    count ++;
    break;
    case 6:                                           // Read Dataset MSL Altitude
    alt = data;
    LSLshifter = 0;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 7:
    LSLshifter = LSLshifter+8;
    tmp32 = data;
    tmp32 = tmp32<<LSLshifter;
    alt = alt | tmp32;
    if (LSLshifter == 24){alt = alt/100; pstep++;}    // altitude in meter
    chkA = chkA + data;
    count ++;
    break;
    case 8:                                           // Read Dataset Ground Speed
    grspeed = data;
    LSLshifter = 0;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 9:
    LSLshifter = LSLshifter+8;
    tmp32 = data;
    tmp32 = tmp32<<LSLshifter;
    grspeed = grspeed | tmp32;
    if (LSLshifter == 24) pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 10:                                          // Read Dataset Heading
    grcourse = data;
    LSLshifter = 0;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 11:
    LSLshifter = LSLshifter+8;
    tmp32 = data;
    tmp32 = tmp32<<LSLshifter;
    grcourse = grcourse | tmp32;
    if (LSLshifter == 24) pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 12:                                          // Read number of satellites in view
    satn = data;
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 13:                                          // Read Fix Type
    fixtype = data;                                   // FIX_NONE = 1, FIX_2D = 2, FIX_3D = 3, FIX_2D_SBAS = 6, FIX_3D_SBAS = 7
    pstep++;
    chkA = chkA + data;
    count ++;
    break;
    case 14:                                          // Wait for cheksum A
    if (count == 33){                                 // 33 = 0x21
      if (chkA == data) pstep++;                      // ChecksumA reached. Correct? than go on
       else pstep = 0;                                // Error?
    }else{
      chkA = chkA + data;
      count ++;}
    break;
    case 15:                                          // Dataset RDY !! Cheksum B omitted, ChkA was OK
		if (fixtype > 1) f.GPS_FIX = true;
		else f.GPS_FIX = false;
	  if (startbyte == 0xd0){                           // We are dealing with old and faulty binary protocol here
      lat = lat * 10;                                 // so we have to multiply by 10 lat and lon
		  lon = lon * 10;}
		GPS_coord[LAT] = lat;                             // GPS_read[LAT] = lat; GPS_read[LON] = lon;
		GPS_coord[LON] = lon;
		GPS_altitude = alt;                               // i2c_dataset.altitude = alt;
    GPS_speed = grspeed;                              // i2c_dataset.ground_speed = grspeed;
    GPS_ground_course = grcourse;                     // i2c_dataset.ground_course = grcourse;
    GPS_numSat = satn;                                // i2c_dataset.status.numsats = satn;
    GPS_Present = 1;                                  // Naze: Show GPS is working
		parsed = true;                                    // RDY
    pstep = 0;                                        // Do nothing
    break;
   }
 return parsed;
}
/*  if (fixtype==1){                                  // No Fix
    i2c_dataset.status.gps2dfix = 0;
    i2c_dataset.status.gps3dfix = 0;}
    if (fixtype==2){                                  // GPS 2D Fix
    i2c_dataset.status.gps2dfix = 1;
    i2c_dataset.status.gps3dfix = 0;}
    if (fixtype>2){                                   // GPS 3D Fix OR DGPS Fix
    i2c_dataset.status.gps2dfix = 1;
    i2c_dataset.status.gps3dfix = 1;} */


/* This is a light implementation of a GPS frame decoding
   This should work with most of modern GPS devices configured to output NMEA frames.
   It assumes there are some NMEA GGA frames to decode on the serial bus
   Here we use only the following data :
     - latitude
     - longitude
     - GPS fix is/is not ok
     - GPS num sat (4 is enough to be +/- reliable)
     // added by Mis
     - GPS altitude (for OSD displaying)
     - GPS speed (for OSD displaying)
*/
#define FRAME_GGA  1
#define FRAME_RMC  2
  
static bool GPS_NMEA_newFrame(char c)              //mwii1320
{
    uint8_t frameOK = 0;
    static uint8_t param = 0, offset = 0, parity = 0;
    static char string[15];
    static uint8_t checksum_param, frame = 0;
  
    if (c == '$') {
      param = 0; offset = 0; parity = 0;
    } else if (c == ',' || c == '*') {
      string[offset] = 0;
      if (param == 0) { //frame identification
        frame = 0;
        if (string[0] == 'G' && string[1] == 'P' && string[2] == 'G' && string[3] == 'G' && string[4] == 'A') frame = FRAME_GGA;
        if (string[0] == 'G' && string[1] == 'P' && string[2] == 'R' && string[3] == 'M' && string[4] == 'C') frame = FRAME_RMC;
      } else if (frame == FRAME_GGA) {
        if      (param == 2)                     {GPS_coord[LAT] = GPS_coord_to_degrees(string);}
        else if (param == 3 && string[0] == 'S') GPS_coord[LAT] = -GPS_coord[LAT];
        else if (param == 4)                     {GPS_coord[LON] = GPS_coord_to_degrees(string);}
        else if (param == 5 && string[0] == 'W') GPS_coord[LON] = -GPS_coord[LON];
        else if (param == 6)                     {f.GPS_FIX = (string[0]  > '0');}
        else if (param == 7)                     {GPS_numSat = grab_fields(string,0);}
        else if (param == 9)                     {GPS_altitude = grab_fields(string,0);}  // altitude in meters added by Mis
      } else if (frame == FRAME_RMC) {
        if      (param == 7)                     {GPS_speed = ((uint32_t)grab_fields(string,1)*5144L)/1000L;}  //gps speed in cm/s will be used for navigation
        else if (param == 8)                     {GPS_ground_course = grab_fields(string,1); }                 //ground course deg*10 
      }
      param++; offset = 0;
      if (c == '*') checksum_param=1;
      else parity ^= c;
    } else if (c == '\r' || c == '\n') {
      if (checksum_param) { //parity checksum
        uint8_t checksum = hex_c(string[0]);
        checksum <<= 4;
        checksum += hex_c(string[1]);
        if (checksum == parity) frameOK = 1;
      }
      checksum_param=0;
    } else {
       if (offset < 15) string[offset++] = c;
       if (!checksum_param) parity ^= c;
    }
    if (frame) GPS_Present = 1;
    return frameOK && (frame==FRAME_GGA);
}

// UBX support
typedef struct {
    uint8_t preamble1;
    uint8_t preamble2;
    uint8_t msg_class;
    uint8_t msg_id;
    uint16_t length;
} ubx_header;

typedef struct {
    uint32_t time;              // GPS msToW
    int32_t longitude;
    int32_t latitude;
    int32_t altitude_ellipsoid;
    int32_t altitude_msl;
    uint32_t horizontal_accuracy;
    uint32_t vertical_accuracy;
} ubx_nav_posllh;

typedef struct {
    uint32_t time;              // GPS msToW
    uint8_t fix_type;
    uint8_t fix_status;
    uint8_t differential_status;
    uint8_t res;
    uint32_t time_to_first_fix;
    uint32_t uptime;            // milliseconds
} ubx_nav_status;

typedef struct {
    uint32_t time;
    int32_t time_nsec;
    int16_t week;
    uint8_t fix_type;
    uint8_t fix_status;
    int32_t ecef_x;
    int32_t ecef_y;
    int32_t ecef_z;
    uint32_t position_accuracy_3d;
    int32_t ecef_x_velocity;
    int32_t ecef_y_velocity;
    int32_t ecef_z_velocity;
    uint32_t speed_accuracy;
    uint16_t position_DOP;
    uint8_t res;
    uint8_t satellites;
    uint32_t res2;
} ubx_nav_solution;

typedef struct {
    uint32_t time;              // GPS msToW
    int32_t ned_north;
    int32_t ned_east;
    int32_t ned_down;
    uint32_t speed_3d;
    uint32_t speed_2d;
    int32_t heading_2d;
    uint32_t speed_accuracy;
    uint32_t heading_accuracy;
} ubx_nav_velned;

enum {
    PREAMBLE1 = 0xb5,
    PREAMBLE2 = 0x62,
    CLASS_NAV = 0x01,
    CLASS_ACK = 0x05,
    CLASS_CFG = 0x06,
    MSG_ACK_NACK = 0x00,
    MSG_ACK_ACK = 0x01,
    MSG_POSLLH = 0x2,
    MSG_STATUS = 0x3,
    MSG_SOL = 0x6,
    MSG_VELNED = 0x12,
    MSG_CFG_PRT = 0x00,
    MSG_CFG_RATE = 0x08,
    MSG_CFG_SET_RATE = 0x01,
    MSG_CFG_NAV_SETTINGS = 0x24
} ubs_protocol_bytes;

enum {
    FIX_NONE = 0,
    FIX_DEAD_RECKONING = 1,
    FIX_2D = 2,
    FIX_3D = 3,
    FIX_GPS_DEAD_RECKONING = 4,
    FIX_TIME = 5
} ubs_nav_fix_type;

enum {
    NAV_STATUS_FIX_VALID = 1
} ubx_nav_status_bits;

// Packet checksum accumulators
static uint8_t _ck_a;
static uint8_t _ck_b;

// State machine state
static uint8_t _step;
static uint8_t _msg_id;
static uint16_t _payload_length;
static uint16_t _payload_counter;

static bool next_fix;
static uint8_t _class;

// do we have new position information?
static bool _new_position;

// do we have new speed information?
static bool _new_speed;

static uint8_t _disable_counter; // Compilerwarning unused Variable

// Receive buffer
static union {
    ubx_nav_posllh posllh;
    ubx_nav_status status;
    ubx_nav_solution solution;
    ubx_nav_velned velned;
    uint8_t bytes[64];
} _buffer;

void _update_checksum(uint8_t *data, uint8_t len, uint8_t *ck_a, uint8_t *ck_b)
{
    while (len--) {
        *ck_a += *data;
        *ck_b += *ck_a;
        data++;
    }
}

static bool GPS_UBLOX_newFrame(uint8_t data)
{
    bool parsed = false;

    switch (_step) {
        case 1:
            if (PREAMBLE2 == data) {
                _step++;
                break;
            }
            _step = 0;
        case 0:
            if (PREAMBLE1 == data)
                _step++;
            break;
        case 2:
            _step++;
            _class = data;
            _ck_b = _ck_a = data;   // reset the checksum accumulators
            break;
        case 3:
            _step++;
            _ck_b += (_ck_a += data);       // checksum byte
            _msg_id = data;
            break;
        case 4:
            _step++;
            _ck_b += (_ck_a += data);       // checksum byte
            _payload_length = data; // payload length low byte
            break;
        case 5:
            _step++;
            _ck_b += (_ck_a += data);       // checksum byte
            _payload_length += (uint16_t) (data << 8);
            if (_payload_length > 512) {
                _payload_length = 0;
                _step = 0;
            }
            _payload_counter = 0;   // prepare to receive payload
            break;
        case 6:
            _ck_b += (_ck_a += data);       // checksum byte
            if (_payload_counter < sizeof(_buffer)) {
                _buffer.bytes[_payload_counter] = data;
            }
            if (++_payload_counter == _payload_length)
                _step++;
            break;
        case 7:
            _step++;
            if (_ck_a != data)
                _step = 0;          // bad checksum
            break;
        case 8:
            _step = 0;
            if (_ck_b != data)
                break;              // bad checksum
            GPS_Present = 1;
            if (UBLOX_parse_gps()) {
                parsed = true;
            }
    }                           //end switch
    return parsed;
}

static bool UBLOX_parse_gps(void)
{
    switch (_msg_id) {
    case MSG_POSLLH:
        //i2c_dataset.time                = _buffer.posllh.time;
        GPS_coord[LON] = _buffer.posllh.longitude;
        GPS_coord[LAT] = _buffer.posllh.latitude;
        GPS_altitude = _buffer.posllh.altitude_msl / 10 / 100;  //alt in m
        f.GPS_FIX = next_fix;
        _new_position = true;
        break;
    case MSG_STATUS:
        next_fix = (_buffer.status.fix_status & NAV_STATUS_FIX_VALID) && (_buffer.status.fix_type == FIX_3D);
        if (!next_fix)
            f.GPS_FIX = false;
        break;
    case MSG_SOL:
        next_fix = (_buffer.solution.fix_status & NAV_STATUS_FIX_VALID) && (_buffer.solution.fix_type == FIX_3D);
        if (!next_fix)
            f.GPS_FIX = false;
        GPS_numSat = _buffer.solution.satellites;
        // GPS_hdop                        = _buffer.solution.position_DOP;
        // debug[3] = GPS_hdop;
        break;
    case MSG_VELNED:
        // speed_3d                        = _buffer.velned.speed_3d;  // cm/s
        GPS_speed = _buffer.velned.speed_2d;    // cm/s
        GPS_ground_course = (uint16_t) (_buffer.velned.heading_2d / 10000);     // Heading 2D deg * 100000 rescaled to deg * 10
        _new_speed = true;
        break;
    default:
        return false;
    }

    // we only return true when we get new position and speed data
    // this ensures we don't use stale data
    if (_new_position && _new_speed) {
        _new_speed = _new_position = false;
        return true;
    }
    return false;
}
