#include "board.h"
#include "mw.h"

// October 2012     V2.1-dev

flags_t f;
int16_t debug[4];
uint8_t toggleBeep = 0;
uint32_t currentTime = 0;
uint32_t previousTime = 0;
uint16_t cycleTime = 0;         // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
int16_t headFreeModeHold;

int16_t annex650_overrun_count = 0;
uint8_t vbat;                   // battery voltage in 0.1V steps
int16_t telemTemperature1;      // gyro sensor temperature

int16_t failsafeCnt = 0;
int16_t rcData[8] = { 1502, 1502, 1502, 1502, 1502, 1502, 1502, 1502 }; // interval [1000;2000]
int16_t rcCommand[4];           // interval [1000;2000] for THROTTLE and [-500;+500] for ROLL/PITCH/YAW 
int16_t lookupPitchRollRC[6];   // lookup table for expo & RC rate PITCH+ROLL
int16_t lookupThrottleRC[11];   // lookup table for expo & mid THROTTLE
rcReadRawDataPtr rcReadRawFunc = NULL;  // receive data from default (pwm/ppm) or additional (spek/sbus/?? receiver drivers)
uint8_t dynP8[3], dynI8[3], dynD8[3];
uint8_t rcOptions[CHECKBOXITEMS];
int16_t axisPID[3];

// **********************
// GPS
// **********************
int32_t GPS_coord[2];
int32_t GPS_home[2];
int32_t GPS_hold[2];
uint8_t GPS_numSat;
uint16_t GPS_distanceToHome;        // distance to home point in meters
int16_t GPS_directionToHome;        // direction to home or hol point in degrees
uint16_t GPS_altitude, GPS_speed;   // altitude in 0.1m and speed in 0.1m/s
uint8_t GPS_update = 0;             // it's a binary toogle to distinct a GPS position update
int16_t GPS_angle[2] = { 0, 0 };    // it's the angles that must be applied for GPS correction
uint16_t GPS_ground_course = 0;     // degrees * 10
uint8_t GPS_Present = 0;            // Checksum from Gps serial
uint8_t GPS_Enable = 0;
int16_t nav[2];
int16_t nav_rated[2];               // Adding a rate controller to the navigation to make it smoother
int8_t nav_mode = NAV_MODE_NONE;    // Navigation mode


// Automatic ACC Offset Calibration
uint16_t InflightcalibratingA = 0;
int16_t AccInflightCalibrationArmed;
uint16_t AccInflightCalibrationMeasurementDone = 0;
uint16_t AccInflightCalibrationSavetoEEProm = 0;
uint16_t AccInflightCalibrationActive = 0;

// Battery monitoring stuff
uint8_t batteryCellCount = 3;       // cell count
uint16_t batteryWarningVoltage;     // annoying buzzer after this one, battery ready to be dead

// Crashpilot LED Inverter stuff
void LD0_OFF(void)
{
  if (cfg.led_invert == 0) LED0_OFF;
	if (cfg.led_invert == 1) LED0_ON;
}
void LD0_ON(void)
{
  if (cfg.led_invert == 0) LED0_ON;
	if (cfg.led_invert == 1) LED0_OFF;
}
void LD1_OFF(void)
{
  if (cfg.led_invert == 0) LED1_OFF;
	if (cfg.led_invert == 1) LED1_ON;
}
void LD1_ON(void)
{
  if (cfg.led_invert == 0) LED1_ON;
	if (cfg.led_invert == 1) LED1_OFF;
}

void blinkLED(uint8_t num, uint8_t wait, uint8_t repeat)
{
    uint8_t i, r;

    for (r = 0; r < repeat; r++) {
        for (i = 0; i < num; i++) {
            LED0_TOGGLE;            // switch LEDPIN state
            BEEP_ON;
            delay(wait);
            BEEP_OFF;
        }
        delay(60);
    }
}

#define BREAKPOINT 1500

// this code is executed at each loop and won't interfere with control loop if it lasts less than 650 microseconds
void annexCode(void)                                 // Crashpilot: execute in computeIMU
{
    static uint32_t calibratedAccTime;
    uint16_t tmp, tmp2;
    static uint8_t buzzerFreq;  //delay between buzzer ring
    static uint8_t vbatTimer = 0;
    uint8_t axis, prop1, prop2;
    static uint8_t ind = 0;
    uint16_t vbatRaw = 0;
    static uint16_t vbatRawArray[8];
    uint8_t i;

    // PITCH & ROLL only dynamic PID adjustemnt,  depending on throttle value
    if (rcData[THROTTLE] < BREAKPOINT) {
        prop2 = 100;
    } else {
        if (rcData[THROTTLE] < 2000) {
            prop2 = 100 - (uint16_t) cfg.dynThrPID * (rcData[THROTTLE] - BREAKPOINT) / (2000 - BREAKPOINT);
        } else {
            prop2 = 100 - cfg.dynThrPID;
        }
    }

    for (axis = 0; axis < 3; axis++) {
        tmp = min(abs(rcData[axis] - cfg.midrc), 500);
        if (axis != 2) {        // ROLL & PITCH
            if (cfg.deadband) {
                if (tmp > cfg.deadband) {
                    tmp -= cfg.deadband;
                } else {
                    tmp = 0;
                }
            }

            tmp2 = tmp / 100;
            rcCommand[axis] = lookupPitchRollRC[tmp2] + (tmp - tmp2 * 100) * (lookupPitchRollRC[tmp2 + 1] - lookupPitchRollRC[tmp2]) / 100;
            prop1 = 100 - (uint16_t) cfg.rollPitchRate * tmp / 500;
            prop1 = (uint16_t) prop1 *prop2 / 100;
        } else {                // YAW
            if (cfg.yawdeadband) {
                if (tmp > cfg.yawdeadband) {
                    tmp -= cfg.yawdeadband;
                } else {
                    tmp = 0;
                }
            }
            rcCommand[axis] = tmp;
            prop1 = 100 - (uint16_t) cfg.yawRate * tmp / 500;
        }
        dynP8[axis] = (uint16_t) cfg.P8[axis] * prop1 / 100;
        dynD8[axis] = (uint16_t) cfg.D8[axis] * prop1 / 100;
        if (rcData[axis] < cfg.midrc)
            rcCommand[axis] = -rcCommand[axis];
    }

    tmp = constrain(rcData[THROTTLE], cfg.mincheck, 2000);
    tmp = (uint32_t) (tmp - cfg.mincheck) * 1000 / (2000 - cfg.mincheck);       // [MINCHECK;2000] -> [0;1000]
    tmp2 = tmp / 100;
    rcCommand[THROTTLE] = lookupThrottleRC[tmp2] + (tmp - tmp2 * 100) * (lookupThrottleRC[tmp2 + 1] - lookupThrottleRC[tmp2]) / 100;    // [0;1000] -> expo -> [MINTHROTTLE;MAXTHROTTLE]

    if(f.HEADFREE_MODE) {
        float radDiff = (heading - headFreeModeHold) * M_PI / 180.0f;
        float cosDiff = cosf(radDiff);
        float sinDiff = sinf(radDiff);
        int16_t rcCommand_PITCH = rcCommand[PITCH] * cosDiff + rcCommand[ROLL] * sinDiff;
        rcCommand[ROLL] = rcCommand[ROLL] * cosDiff - rcCommand[PITCH] * sinDiff;
        rcCommand[PITCH] = rcCommand_PITCH;
    }

    if (feature(FEATURE_VBAT)) {
        if (!(++vbatTimer % VBATFREQ)) {
            vbatRawArray[(ind++) % 8] = adcGetChannel(ADC_BATTERY);
            for (i = 0; i < 8; i++)
                vbatRaw += vbatRawArray[i];
            vbat = batteryAdcToVoltage(vbatRaw / 8);
        }
        if ((vbat > batteryWarningVoltage) || (vbat < cfg.vbatmincellvoltage)) { // VBAT ok, buzzer off
            buzzerFreq = 0;
        } else
            buzzerFreq = 4;     // low battery
    }

    buzzer(buzzerFreq);         // external buzzer routine that handles buzzer events globally now

    if ((calibratingA > 0 && sensors(SENSOR_ACC)) || (calibratingG > 0)) {      // Calibration phasis
        LED0_TOGGLE;
    } else {
        if (f.ACC_CALIBRATED)
            LD0_OFF();    // Crashpilot LED0_OFF;
        if (f.ARMED)
            LD0_ON();     // Crashpilot LED0_ON;
        // This will switch to/from 9600 or 115200 baud depending on state. Of course, it should only do it on changes.
        if (feature(FEATURE_TELEMETRY))
            initTelemetry(f.ARMED);
    }

#ifdef LEDRING
    if (feature(FEATURE_LED_RING)) {
        static uint32_t LEDTime;
        if ((int32_t)(currentTime - LEDTime) >= 0) {
            LEDTime = currentTime + 50000;
            ledringState();
        }
    }
#endif

    if ((int32_t)(currentTime - calibratedAccTime) >= 0) {
        if (!f.SMALL_ANGLES_25) {
            f.ACC_CALIBRATED = 0; // the multi uses ACC and is not calibrated or is too much inclinated
            LED0_TOGGLE;
            calibratedAccTime = currentTime + 500000;
        } else {
            f.ACC_CALIBRATED = 1;
        }
    }

    serialCom();

    if (sensors(SENSOR_GPS)){                                 // Crashpilot Show number of Sats by Blinking
        static uint32_t Timebase;
			  static uint8_t  cnt,blinkcase,blinkblock;
			  if (GPS_numSat < 5) blinkcase = 0;
			  if (GPS_numSat >= 5 && blinkcase == 0){
					blinkcase = 1;
					blinkblock = GPS_numSat;
					Timebase = 0;
        }
				if (currentTime >= Timebase){
				  Timebase = currentTime + 10000;                     // Timebase 10 ms
					LD1_OFF();                                          // LED1_OFF;
				  switch(blinkcase){
				  	case 0:
					  break;
					  case 1:
 						  cnt = 30;                                       // 300 ms on
						  blinkcase++;
				    break;
					  case 2:
              LED1_TOGGLE;
              cnt--;
              if (cnt == 0) blinkcase++;
				    break;
					  case 3:
 						  cnt = 70;                                       // 700 ms off
						  blinkcase++;
				    break;
					  case 4:
              cnt--;
              if (cnt == 0) blinkcase++;
				    break;
					  case 5:
							blinkblock--;
              if (blinkblock == 0) blinkcase++;
						    else blinkcase = 1;
				    break;
					  case 6:
 	  					cnt = 150;                                      // 1.5s Break before next flickerblock
						  blinkcase++;
				    break;
					  case 7:
						  cnt--;
              if (cnt == 0) blinkcase=0;
				    break;
				  }
			 }
    }

    // Read out gyro temperature. can use it for something somewhere. maybe get MCU temperature instead? lots of fun possibilities.
    if (gyro.temperature)
        gyro.temperature(&telemTemperature1);
    else {
        // TODO MCU temp
    }
} // END OF annexCode 

uint16_t pwmReadRawRC(uint8_t chan)
{
    uint16_t data;

    data = pwmRead(cfg.rcmap[chan]);
    if (data < 750 || data > 2250)
        data = cfg.midrc;

    return data;
}

void computeRC(void)
{
    static int16_t rcData4Values[8][4], rcDataMean[8];
    static uint8_t rc4ValuesIndex = 0;
    uint8_t chan, a;

    rc4ValuesIndex++;
    for (chan = 0; chan < 8; chan++) {
        rcData4Values[chan][rc4ValuesIndex % 4] = rcReadRawFunc(chan);
        rcDataMean[chan] = 0;
        for (a = 0; a < 4; a++)
            rcDataMean[chan] += rcData4Values[chan][a];

        rcDataMean[chan] = (rcDataMean[chan] + 2) / 4;
        if (rcDataMean[chan] < rcData[chan] - 3)
            rcData[chan] = rcDataMean[chan] + 2;
        if (rcDataMean[chan] > rcData[chan] + 3)
            rcData[chan] = rcDataMean[chan] - 2;
    }
}

void pass(void)
{
	  static uint32_t blinktime;
    uint32_t timetmp;
  	if (spektrumFrameComplete()) computeRC();
    if (!feature(FEATURE_SPEKTRUM)) computeRC();
	  writeAllMotors(rcData[THROTTLE]);
	  serialCom();
  	timetmp = micros();
	  if ((int32_t)(timetmp - blinktime) >= 0) {
      blinktime = timetmp + 150000;
 	    LED1_TOGGLE;
	    LED0_TOGGLE;
		}
}

void loop(void)
{
	  #define F_CUT 20.0f                            // pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624
    #define RC 0.5f / (M_PI * F_CUT)               // pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624
 	  static float lastDTerm[3] = { 0.0, 0.0, 0.0};  // pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624

    static uint8_t rcDelayCommand;                 // this indicates the number of time (multiple of RC measurement at 50Hz) the sticks must be maintained to run or switch off motors
    uint8_t axis, i;
    float error, errorAngle;
    float delta, deltaSum;
    float PTerm, ITerm, PTermACC, ITermACC = 0, PTermGYRO = 0, ITermGYRO = 0, DTerm;
    static int16_t lastGyro[3] = { 0, 0, 0 };
    static float delta1[3], delta2[3];
    static float errorGyroI[3] = { 0, 0, 0 };
    static float errorAngleI[2] = { 0, 0 };
    static uint32_t rcTime = 0;
    static int16_t initialThrottleHold;
    uint16_t auxState = 0;
    float prop;

    static uint32_t AltRCTimer0;               // Crashpilot
    static uint8_t  HightChange=0;             // Crashpilot
    static int32_t  AltHold8;                  // Crashpilot
    static uint8_t  ThrFstTimeCenter;          // Crashpilot
    static uint8_t  toggle;                    // Crashpilot
    int16_t         VirtualThrottle;           // Crashpilot
		static uint32_t PosHoldBlindTimer=0;       // Crashpilot
		static uint8_t  PSholdChange=0;            // Crashpilot
    static uint32_t AutolandGeneralTimer;      // Crashpilot AUTOLAND
    static uint8_t  AutolandState;             // Crashpilot AUTOLAND
    static int16_t  LastAltThrottle;           // Crashpilot AUTOLAND
   	static uint32_t RTLGeneralTimer;           // Crashpilot: RTL
		static uint8_t  RTLstate;                  // Crashpilot: RTL
		static uint8_t  PHminSat;                  // Crashpilot: General GPS
		
    // this will return false if spektrum is disabled. shrug.
    if (spektrumFrameComplete())
        computeRC();

    if ((int32_t)(currentTime - rcTime) >= 0) { // 50Hz
        rcTime = currentTime + 20000;
        // TODO clean this up. computeRC should handle this check
        if (!feature(FEATURE_SPEKTRUM))
            computeRC();

        if ((rcData[THROTTLE] < cfg.mincheck) && AutolandState == 0 ){ // Crashpilot
            errorGyroI[ROLL] = 0;
            errorGyroI[PITCH] = 0;
            errorGyroI[YAW] = 0;
            errorAngleI[ROLL] = 0;
            errorAngleI[PITCH] = 0;
            rcDelayCommand++;
            if (rcData[YAW] < cfg.mincheck && rcData[PITCH] < cfg.mincheck && !f.ARMED) {
                if (rcDelayCommand == 20) {
                    calibratingG = 1000;
                    if (feature(FEATURE_GPS))
                        GPS_reset_home_position();
                }
            } else if (feature(FEATURE_INFLIGHT_ACC_CAL) && (!f.ARMED && rcData[YAW] < cfg.mincheck && rcData[PITCH] > cfg.maxcheck && rcData[ROLL] > cfg.maxcheck)) {
                if (rcDelayCommand == 20) {
                    if (AccInflightCalibrationMeasurementDone) {        // trigger saving into eeprom after landing
                        AccInflightCalibrationMeasurementDone = 0;
                        AccInflightCalibrationSavetoEEProm = 1;
                    } else {
                        AccInflightCalibrationArmed = !AccInflightCalibrationArmed;
                        if (AccInflightCalibrationArmed) {
                            toggleBeep = 2;
                        } else {
                            toggleBeep = 3;
                        }
                    }
                }
            } else if (cfg.activate[BOXARM] > 0) {
                if (rcOptions[BOXARM] && f.OK_TO_ARM) {
                    // TODO: feature(FEATURE_FAILSAFE) && failsafeCnt == 0
                    f.ARMED = 1;
                    headFreeModeHold = heading;
                } else if (f.ARMED)
                    f.ARMED = 0;
                rcDelayCommand = 0;
            } else if ((rcData[YAW] < cfg.mincheck || (cfg.retarded_arm == 1 && rcData[ROLL] < cfg.mincheck)) && f.ARMED) {
                if (rcDelayCommand == 20)
                    f.ARMED = 0;  // rcDelayCommand = 20 => 20x20ms = 0.4s = time to wait for a specific RC command to be acknowledged
            } else if ((rcData[YAW] > cfg.maxcheck || (rcData[ROLL] > cfg.maxcheck && cfg.retarded_arm == 1)) && rcData[PITCH] < cfg.maxcheck && !f.ARMED && calibratingG == 0 && f.ACC_CALIBRATED) {
                if (rcDelayCommand == 20) {
                    f.ARMED = 1;
                    headFreeModeHold = heading;
                }
            } else
                rcDelayCommand = 0;
        } else if (rcData[THROTTLE] > cfg.maxcheck && !f.ARMED) {
            if (rcData[YAW] < cfg.mincheck && rcData[PITCH] < cfg.mincheck) {   // throttle=max, yaw=left, pitch=min
                if (rcDelayCommand == 20)
                    calibratingA = 400;
                rcDelayCommand++;
            } else if (rcData[YAW] > cfg.maxcheck && rcData[PITCH] < cfg.mincheck) {    // throttle=max, yaw=right, pitch=min
                if (rcDelayCommand == 20)
                    f.CALIBRATE_MAG = 1;   // MAG calibration request
                rcDelayCommand++;
            } else if (rcData[PITCH] > cfg.maxcheck) {
                cfg.angleTrim[PITCH] += 2;
                writeParams(1);
#ifdef LEDRING
                if (feature(FEATURE_LED_RING))
                    ledringBlink();
#endif
            } else if (rcData[PITCH] < cfg.mincheck) {
                cfg.angleTrim[PITCH] -= 2;
                writeParams(1);
#ifdef LEDRING
                if (feature(FEATURE_LED_RING))
                    ledringBlink();
#endif
            } else if (rcData[ROLL] > cfg.maxcheck) {
                cfg.angleTrim[ROLL] += 2;
                writeParams(1);
#ifdef LEDRING
                if (feature(FEATURE_LED_RING))
                    ledringBlink();
#endif
            } else if (rcData[ROLL] < cfg.mincheck) {
                cfg.angleTrim[ROLL] -= 2;
                writeParams(1);
#ifdef LEDRING
                if (feature(FEATURE_LED_RING))
                    ledringBlink();
#endif
            } else {
                rcDelayCommand = 0;
            }
        }

        if (feature(FEATURE_INFLIGHT_ACC_CAL)) {
            if (AccInflightCalibrationArmed && f.ARMED && rcData[THROTTLE] > cfg.mincheck && !rcOptions[BOXARM]) {   // Copter is airborne and you are turning it off via boxarm : start measurement
                InflightcalibratingA = 50;
                AccInflightCalibrationArmed = 0;
            }
            if (rcOptions[BOXPASSTHRU]) {      // Use the Passthru Option to activate : Passthru = TRUE Meausrement started, Land and passtrhu = 0 measurement stored
                if (!AccInflightCalibrationActive && !AccInflightCalibrationMeasurementDone)
                    InflightcalibratingA = 50;
            } else if (AccInflightCalibrationMeasurementDone && !f.ARMED) {
                AccInflightCalibrationMeasurementDone = 0;
                AccInflightCalibrationSavetoEEProm = 1;
            }
        }

        for (i = 0; i < 4; i++)
            auxState |= (rcData[AUX1 + i] < 1300) << (3 * i) | (1300 < rcData[AUX1 + i] && rcData[AUX1 + i] < 1700) << (3 * i + 1) | (rcData[AUX1 + i] > 1700) << (3 * i + 2);
        for (i = 0; i < CHECKBOXITEMS; i++)
            rcOptions[i] = (auxState & cfg.activate[i]) > 0;

        /////////////
			  PHminSat = 7;                                            // Don't forget to set PH Minsats here!!
        /////////////
// Failsafe routine Modified & Moved here by Crashpilot
				// 4 Cases:
				// 1: No Sensors -> Do the old stuff
				// 2: GPS -> Do RTL and do the old stuff
				// 3: Baro -> Do Baro Autoland
				// 4: BARO & GPS -> Do full feature RTL				
        if (feature(FEATURE_FAILSAFE) && f.ARMED)                // Only check Failsafe if copter is armed
				{
					if (failsafeCnt > (5 * cfg.failsafe_delay)){           // OMG we have a failsafe
				    rcData[ROLL] = cfg.midrc;                            // Center Sticks
					  rcData[PITCH] = cfg.midrc;                           // Center Sticks
					  rcData[YAW] = cfg.midrc;                             // Center Sticks
						rcOptions[BOXPASSTHRU] = 0;                          // Passthru off

 			      if (!sensors(SENSOR_GPS) && !sensors(SENSOR_BARO)){  // No Sensors just go down
					    rcData[THROTTLE] = cfg.failsafe_throttle;
	            if (failsafeCnt > 5 * (cfg.failsafe_delay + cfg.failsafe_off_delay)){ // Turn OFF motors Time in 0.1sec
                f.ARMED = 0; 
                f.OK_TO_ARM = 0;
              }
						}
				    if (sensors(SENSOR_GPS) && !sensors(SENSOR_BARO)){   // Just GPS. We initiate RTL not caring if homepos is set, because he is going down anyway
              rcOptions[BOXMAG]      = 1;                        // MAG ON
              rcOptions[BOXHEADFREE] = 0;                        // HeadFree off	
					    rcData[THROTTLE] = cfg.failsafe_throttle;
		          PHminSat = 5;                                      // Sloppy PH is sufficient
              if (f.GPS_FIX_HOME) rcOptions[BOXGPSHOME]  = 1;    // Do RTL
							else rcOptions[BOXGPSHOLD]  = 1;                   // OMG we have no Homepos, turn on PH just in case...
	            if (failsafeCnt > 5 * (cfg.failsafe_delay + cfg.failsafe_off_delay)){ // Turn OFF motors Time in 0.1sec
                f.ARMED = 0; 
                f.OK_TO_ARM = 0;
              }
						}
 			      if (!sensors(SENSOR_GPS) && sensors(SENSOR_BARO)){   // Just Baro. Do Autoland and drift into next tree
							rcOptions[BOXBARO]     = 1;                        // Baro On
              rcData[THROTTLE] = cfg.mincheck-10;                // Do Autoland turns off motors
						}							
			      if (sensors(SENSOR_GPS) && sensors(SENSOR_BARO)){    // GPS & Baro. Do RTH & Autoland. If Homepos not set just Autoland & try PH
              if (f.GPS_FIX_HOME) rcOptions[BOXGPSHOME]  = 1;    // Do RTL+Autoland+MotorOFF
							else{                                              // OMG we have no Homepos - just do Autoland, turn on PH just in case...
		            PHminSat = 5;                                    // Sloppy PH is sufficient
                rcOptions[BOXMAG]      = 1;                      // MAG ON
                rcOptions[BOXHEADFREE] = 0;                      // HeadFree off	
                rcOptions[BOXBARO]     = 1;                      // Baro On              
                rcOptions[BOXGPSHOLD]  = 1;                      // Pos Hold ON
                rcData[THROTTLE] = cfg.mincheck-10;              // Do Autoland turns off motors
							}
						}
					}
				}
        if (feature(FEATURE_FAILSAFE)) failsafeCnt++;            // reset to 0 by pwm / spektrum driver
// Failsafe routine Modified & Moved here by Crashpilot
				
//      SPECIAL RTL Crashpilot
//      Full RTL with Althold+Hightcheck+Autoland+Disarm				
        if (rcOptions[BOXGPSHOME] || rcOptions[BOXGPSHOLD]){     // Switch to Angle mode when GPS is on anyway
         rcOptions[BOXHORIZON] = 0;
				 rcOptions[BOXANGLE]   = 1;
        }
			
				#define RTLsettleTime 2000000                   // 2 sec
				#define RTLClimbRate 64
				if (sensors(SENSOR_GPS) && sensors(SENSOR_BARO) && f.GPS_FIX_HOME && rcOptions[BOXGPSHOME]) {
          rcOptions[BOXBARO]     = 1;                   // Baro On
          rcOptions[BOXMAG]      = 1;                   // MAG ON Idea: Louis
          rcOptions[BOXPASSTHRU] = 0;                   // Passthru off
          rcOptions[BOXHEADFREE] = 0;                   // HeadFree off	
					rcData[THROTTLE] = cfg.midrc;                 // Put throttlestick to middle: Althold
          PHminSat = 5;                                 // Sloppy PH is sufficient

          if (RTLstate == 0) RTLstate = 1;              // Start RTL Sequence if it isn't already running
					if (!f.GPS_FIX || GPS_numSat < 5) RTLstate=0; // Error! No Sat Fix

          switch (RTLstate){
 				  case 0:                                       // Error!! Do landing
					rcData[THROTTLE] = cfg.mincheck-10;           // Put throttlestick to lowest-10
          break;
					case 1:                                       // Engage Althold and PosHold
          rcOptions[BOXGPSHOLD]  = 1;                   // GPS hold On
					rcOptions[BOXGPSHOME]  = 0;                   // Do not engage RTL yet
	        RTLGeneralTimer = currentTime+RTLsettleTime;  // prepare timer 
					RTLstate++;
					break;
					case 2:                                       // Hover certain time
          rcOptions[BOXGPSHOLD]  = 1;                   // GPS hold On
					rcOptions[BOXGPSHOME]  = 0;                   // Do not engage RTL yet
				  if (currentTime > RTLGeneralTimer) RTLstate++;
					break;
					case 3:                                       // Check hight and climb if neccessary
          rcOptions[BOXGPSHOLD]  = 1;                   // GPS hold On
					rcOptions[BOXGPSHOME]  = 0;                   // Do not engage RTL yet
          if (cfg.gps_rtl_minhight == 0) RTLstate++;    // For safety, skip if turned off
					else{
					  if (EstAlt < cfg.gps_rtl_minhight*100) rcData[THROTTLE] = cfg.midrc + cfg.alt_hold_throttle_neutral + RTLClimbRate;
            else RTLstate++;
					}
					break;
					case 4:                                       // Prepare RTL
          rcOptions[BOXGPSHOLD]  = 0;                   // GPS hold OFF
					rcOptions[BOXGPSHOME]  = 1;                   // Engage RTL
					RTLstate++;
					break;
					case 5:                                       // OMG Do the f** RTL now
          rcOptions[BOXGPSHOLD]  = 0;                   // GPS hold OFF
					rcOptions[BOXGPSHOME]  = 1;                   // RTL
					if (nav_mode == NAV_MODE_POSHOLD) RTLstate++; // OK RTL is RDY
					break;
					case 6:                                       // Do Autoland
					rcData[THROTTLE] = cfg.mincheck-10;           // Put throttlestick to lowest-10
          rcOptions[BOXGPSHOLD]  = 1;                   // GPS hold
					rcOptions[BOXGPSHOME]  = 0;                   // RTL OFF
					break;                                        // Repeat forever because Autoland will disarm the thing
				  }				
			  } else RTLstate = 0;                            // No BOXGPSHOME request? Reset Variable
//      SPECIAL RTL Crashpilot END
				
        // note: if FAILSAFE is disable, failsafeCnt > 5*FAILSAVE_DELAY is always false
        if ((rcOptions[BOXANGLE] || (failsafeCnt > 5 * cfg.failsafe_delay)) && (sensors(SENSOR_ACC))) {
            // bumpless transfer to Level mode
            if (!f.ANGLE_MODE) {
                errorAngleI[ROLL] = 0;
                errorAngleI[PITCH] = 0;
                f.ANGLE_MODE = 1;
            }
        } else {
            f.ANGLE_MODE = 0;        // failsave support
        }

        if (rcOptions[BOXHORIZON]) {
            if (!f.HORIZON_MODE) {
                errorAngleI[ROLL] = 0; 
                errorAngleI[PITCH] = 0;
                f.HORIZON_MODE = 1;
            }
        } else {
            f.HORIZON_MODE = 0;
        }

        if ((rcOptions[BOXARM]) == 0)
            f.OK_TO_ARM = 1;
        if (f.ANGLE_MODE || f.HORIZON_MODE) {
            LD1_ON(); //LED1_ON; Crashpilot 
        } else {
						LD1_OFF();                            // Crashpilot LED1_OFF;
        }
				
#ifdef BARO                                        // Crashpilot
		if (!f.ARMED) f.BARO_MODE = 0;
		if (sensors(SENSOR_BARO)) {
      if (rcOptions[BOXBARO]) {
 		    if (((rcData[THROTTLE]) < cfg.mincheck) && AutolandState == 0) AutolandState = 1; // Start Autoland
		    if (((rcData[THROTTLE]) > cfg.mincheck) && AutolandState != 0) AutolandState = 0; // Stop Autoland on Userinput
        if (!f.BARO_MODE) {
          f.BARO_MODE = 1;
          AltHold = EstAlt;
          AltHold8 = AltHold << 3;
          AltRCTimer0 = 0;                                     // The User has 1sec to center throttle after Althold is engaged
          HightChange = 0; ThrFstTimeCenter = 0;
          initialThrottleHold = rcCommand[THROTTLE];}
				}else{
				f.BARO_MODE = 0;
				AutolandState = 0; // No Baroswitch, no Autoland
			}
		}else{
		   AutolandState = 0; // No Baro, no Autoland
		 }
#endif
		
#ifdef  MAG
        if (sensors(SENSOR_MAG)) {
            if (rcOptions[BOXMAG]) {
                if (!f.MAG_MODE) {
                    f.MAG_MODE = 1;
                    magHold = heading;
                }
            } else
                f.MAG_MODE = 0;
            if (rcOptions[BOXHEADFREE]) {
                if (!f.HEADFREE_MODE) {
                    f.HEADFREE_MODE = 1;
                }
            } else {
                f.HEADFREE_MODE = 0;
            }
            if (rcOptions[BOXHEADADJ]) {
                headFreeModeHold = heading; // acquire new heading
            }
        }
#endif
        if (sensors(SENSOR_GPS)) {
            if (f.GPS_FIX && GPS_numSat >= 5) {
                if (rcOptions[BOXGPSHOME] && f.GPS_FIX_HOME) {                              // Crashpilot RTH is possible with 5 Sats for emergency and if homepos is set!
                    if (!f.GPS_HOME_MODE) {
                        f.GPS_HOME_MODE = 1;
                        GPS_set_next_wp(&GPS_home[LAT], &GPS_home[LON]);
                        nav_mode = NAV_MODE_WP;
                    }
                } else f.GPS_HOME_MODE = 0;
                
                if (rcOptions[BOXGPSHOLD] && PSholdChange == 0 && GPS_numSat >= PHminSat) { // Crashpilot Only do poshold with 7 Sats or more
                    if (!f.GPS_HOLD_MODE) {
                        f.GPS_HOLD_MODE = 1;
                        GPS_hold[LAT] = GPS_coord[LAT];
                        GPS_hold[LON] = GPS_coord[LON];
                        GPS_set_next_wp(&GPS_hold[LAT], &GPS_hold[LON]);
                        nav_mode = NAV_MODE_POSHOLD;
                    }
                } else f.GPS_HOLD_MODE = 0;
            } else { // mwii r1315
                f.GPS_HOME_MODE = 0;
                f.GPS_HOLD_MODE = 0;
              }

// Crashpilot poshold Override START
     #define PosHoldBlindTime 300000
     #define BlindZonePitch 30
     #define BlindZoneRoll 30
				
      if (rcOptions[BOXGPSHOLD])
	     {
        if (abs(rcData[PITCH]-cfg.midrc) > BlindZonePitch || abs(rcData[ROLL]-cfg.midrc) > BlindZoneRoll)
	       {
          PosHoldBlindTimer = 0;
          PSholdChange = 1;
         } else {
            if (PSholdChange==1 && PosHoldBlindTimer==0)
              PosHoldBlindTimer = currentTime+PosHoldBlindTime;             // Ph is off and Stick back to neutralzone
           }
         if (currentTime >= PosHoldBlindTimer && PosHoldBlindTimer !=0)
		      {
           PosHoldBlindTimer = 0;
           PSholdChange = 0;
          }
      } else {
         PosHoldBlindTimer = 0;
         PSholdChange = 0;
        }            // Crashpilot poshold Override END				
  } // END of sensors SENSOR_GPS

        if (rcOptions[BOXPASSTHRU]) {
            f.PASSTHRU_MODE = 1;
        } else {
            f.PASSTHRU_MODE = 0;
        }

        if (cfg.mixerConfiguration == MULTITYPE_FLYING_WING || cfg.mixerConfiguration == MULTITYPE_AIRPLANE) {
            f.HEADFREE_MODE = 0;
        }
    } else {                    // not in rc loop
        static int8_t taskOrder = 0;    // never call all function in the same loop, to avoid high delay spikes
        switch (taskOrder++ % 4) {
        case 0:
#ifdef MAG
            if (sensors(SENSOR_MAG))
                Mag_getADC();
#endif
            break;
        case 1:
            break;
        case 2:
            break;
        case 3:
#ifdef SONAR
            if (sensors(SENSOR_SONAR)) {
                Sonar_update();
                debug[2] = sonarAlt;
            }
#endif
            break;
        default:
            taskOrder = 0;
            break;
        }
    }

#ifdef BARO
            if (sensors(SENSOR_BARO)){
                Baro_update();
                getEstimatedAltitude();
						}
#endif

        computeIMU();
        // Measure loop rate just afer reading the sensors MOVED CRASHPILOT
						
#ifdef MPU6050_DMP
        mpu6050DmpLoop();
#endif

#ifdef MAG
        if (sensors(SENSOR_MAG)) {
            if (abs(rcCommand[YAW]) < 70 && f.MAG_MODE) {
                int16_t dif = heading - magHold;
                if (dif <= -180)
                    dif += 360;
                if (dif >= +180)
                    dif -= 360;
                if (f.SMALL_ANGLES_25)
                    rcCommand[YAW] -= dif * cfg.P8[PIDMAG] / 30;    // 18 deg
            } else
                magHold = heading;
        }
#endif
			
#ifdef BARO
#define AutolandRate 64                               // 40cm/sec (64/8*5)
#define HoverTimeBeforeLand 2000000                   // Wait 2 sec in the air for VirtualThrottle to catch up
#define HasLandedTimeCheck 3000000                    // Timeperiod for idlethrottle before shut off
				
		if (sensors(SENSOR_BARO)) {
      if (f.BARO_MODE && f.ARMED)
      {
 		  switch (AutolandState){
        case 0:                                       // No Autoland Do nothing
        break;
				case 1:                                       // Start Althold
				  rcData[THROTTLE] = cfg.midrc;               // Put throttlestick to middle
	        AutolandGeneralTimer = currentTime+HoverTimeBeforeLand; // prepare timer 
				  AutolandState++;
        break;
				case 2:
				  rcData[THROTTLE] = cfg.midrc;               // Put throttlestick to middle: Hover certain time
				  if (currentTime > AutolandGeneralTimer) AutolandState++;
        break;
				case 3:                                       // Start descent initialize Variables
				  rcData[THROTTLE] = cfg.midrc - cfg.alt_hold_throttle_neutral - AutolandRate;
	        AutolandGeneralTimer = 0;
          AutolandState++;
        break;
				case 4:                                       // Check for landing
				  rcData[THROTTLE] = cfg.midrc - cfg.alt_hold_throttle_neutral - AutolandRate;
				  if (LastAltThrottle == cfg.minthrottle && AutolandGeneralTimer == 0 ) AutolandGeneralTimer = currentTime + HasLandedTimeCheck;
					if (LastAltThrottle > cfg.minthrottle) AutolandGeneralTimer = 0; // Reset Timer
          if (AutolandGeneralTimer != 0 && currentTime > AutolandGeneralTimer)  AutolandState++;
        break;
				case 5:                                       // Shut down Copter forever....
				  f.ARMED = 0;
          f.OK_TO_ARM = 0;
        break;
	    }
 			 int16_t thrdiff = rcData[THROTTLE]-cfg.midrc;
       int16_t ABSthrdiff = abs(thrdiff);
       if (ABSthrdiff < cfg.alt_hold_throttle_neutral && ThrFstTimeCenter == 0) ThrFstTimeCenter = 1;
       if (currentTime >= AltRCTimer0){                                                                          // 10 Hz Loop
         AltRCTimer0 = currentTime+100000;
         toggle = toggle^1;                                                                                       // toggle = !toggle;
         if (ThrFstTimeCenter == 1){                                                                              // Begin of ThrFstTimeCenter
          if (ABSthrdiff > cfg.alt_hold_throttle_neutral){
            HightChange = 1;
            initialThrottleHold = initialThrottleHold + BaroP/100;                                                // Adjust Baselinethr by 1% of BaroP
            if (thrdiff >= 0) AltHold8 = AltHold8 + thrdiff - cfg.alt_hold_throttle_neutral;
            if (thrdiff < 0 && toggle == 0) AltHold8 = AltHold8 + thrdiff + cfg.alt_hold_throttle_neutral;
           }else HightChange = 0;}                                                                                // End of ThrFstTimeCenter
        }                                                                                                         // End of X Hz Loop
        AltHold = AltHold8 >> 3;
        VirtualThrottle = initialThrottleHold + BaroP + BaroD - BaroI;
        if (VirtualThrottle <= cfg.minthrottle-300 && HightChange == 1) f.BARO_MODE = 0;                             // has probably landed reset barostuff
        rcCommand[THROTTLE] = constrain(VirtualThrottle,cfg.minthrottle,cfg.maxthrottle);
        LastAltThrottle = rcCommand[THROTTLE];
      }	
	  }
#endif

        if (sensors(SENSOR_GPS)) {
            // Check that we really need to navigate ?
            if ((!f.GPS_HOME_MODE && !f.GPS_HOLD_MODE) || !f.GPS_FIX_HOME) {
                // If not. Reset nav loops and all nav related parameters
                GPS_reset_nav();
            } else {
                float sin_yaw_y = sinf(heading * 0.0174532925f);
                float cos_yaw_x = cosf(heading * 0.0174532925f);
                if (cfg.nav_slew_rate) {
                    nav_rated[LON] += constrain(wrap_18000(nav[LON] - nav_rated[LON]), -cfg.nav_slew_rate, cfg.nav_slew_rate); // TODO check this on uint8
                    nav_rated[LAT] += constrain(wrap_18000(nav[LAT] - nav_rated[LAT]), -cfg.nav_slew_rate, cfg.nav_slew_rate);
                    GPS_angle[ROLL] = (nav_rated[LON] * cos_yaw_x - nav_rated[LAT] * sin_yaw_y) / 10;
                    GPS_angle[PITCH] = (nav_rated[LON] * sin_yaw_y + nav_rated[LAT] * cos_yaw_x) / 10;
                } else {
                    GPS_angle[ROLL] = (nav[LON] * cos_yaw_x - nav[LAT] * sin_yaw_y) / 10;
                    GPS_angle[PITCH] = (nav[LON] * sin_yaw_y + nav[LAT] * cos_yaw_x) / 10;
                }
            }
        }

				// Crashpilot moved it here
        currentTime = micros();
        cycleTime = (int32_t)(currentTime - previousTime);
        previousTime = currentTime;
				
        #define BasePIDtime 3000.0f                                        // Defines the Timebase the PID controller refers to so you can keep your PIDS
				
        // **** PITCH & ROLL & YAW PID ****  
        float dT = cycleTime * 1e-6;				                               // pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624

        prop = max(abs((float)rcCommand[PITCH]), abs((float)rcCommand[ROLL]));  // range [0;500] Crashpilot: prop is float now
        for (axis = 0; axis < 3; axis++) {
            if ((f.ANGLE_MODE || f.HORIZON_MODE) && axis < 2) {            // MODE relying on ACC 50 degrees max inclination

                errorAngle = constrain(2.0f * (float)rcCommand[axis] + (float)GPS_angle[axis], -500.0f, +500.0f) - (float)angle[axis] + (float)cfg.angleTrim[axis];
                errorAngle = errorAngle*(float)cycleTime/BasePIDtime;      // Crashpilot: Include Cylcletime take 3ms as basis. More deltaT more error
  							PTermACC = errorAngle * (float)cfg.P8[PIDLEVEL] / 100.0f;
                PTermACC = constrain(PTermACC, -(float)cfg.D8[PIDLEVEL] * 5.0f, +(float)cfg.D8[PIDLEVEL] * 5.0f);
                errorAngleI[axis] = constrain(errorAngleI[axis] + errorAngle, -10000.0f, +10000.0f); // WindUp
                ITermACC = ((float)errorAngleI[axis] * (float)cfg.I8[PIDLEVEL])/4096.0f;
            }
            if (!f.ANGLE_MODE || axis == 2) {                              // MODE relying on GYRO or YAW axis
                error = (float)rcCommand[axis]*80.0f/(float)cfg.P8[axis];
  							error = error*(float)cycleTime/BasePIDtime;                // Crashpilot: Include Cylcletime take 3ms as basis. More deltaT more error
 			 				  error -= (float)gyroData[axis];
                PTermGYRO = (float)rcCommand[axis];
                errorGyroI[axis] = constrain(errorGyroI[axis] + error, -16000.0f, +16000.0f);
                if (abs(gyroData[axis]) > 640.0f) 
                    errorGyroI[axis] = 0;
                ITermGYRO = (errorGyroI[axis]/125.0f * (float)cfg.I8[axis])/64.0f;
            }
            if (f.HORIZON_MODE && axis < 2) {
                PTerm = (PTermACC * (500.0f - prop) + PTermGYRO * prop) / 500.0f;
                ITerm = (ITermACC * (500.0f - prop) + ITermGYRO * prop) / 500.0f;
            } else {
                if (f.ANGLE_MODE && axis < 2) {
                    PTerm = PTermACC;
                    ITerm = ITermACC;
                } else {
                    PTerm = PTermGYRO;
                    ITerm = ITermGYRO;
                }
            }
            PTerm -= (float)gyroData[axis] * (float)dynP8[axis] / 10.0f / 8.0f;

            delta = (float)gyroData[axis] - (float)lastGyro[axis];
            lastGyro[axis] = gyroData[axis];
						
            deltaSum = delta1[axis] + delta2[axis] + delta;
	          delta2[axis] = delta1[axis];
            delta1[axis] = delta;

            deltaSum = lastDTerm[axis] + (dT / (RC + dT)) * (deltaSum - lastDTerm[axis]); // pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624
            lastDTerm[axis] = deltaSum;						// pt1 element http://www.multiwii.com/forum/viewtopic.php?f=23&t=2624
						
            DTerm = ((float)deltaSum * (float)dynD8[axis])/32.0f;
            axisPID[axis] =  PTerm + ITerm - DTerm;
        }
          mixTable();
          writeServos();
          writeMotors();
}
